///////////////////////////////////////////////////////////////////////////////
//
// Module: logger.v
// Project: BJP1
// Description: 
//              Receives log enteries from input_arbitter and sender modules
//				and send them throgh an Ethernet port.
// 				Each log entery is 64 bit, and each 100 entery will go in one 
//				packet.
//				Every received entery will bufferd first in a little FIFO.
//
// By hamed.ty@gmail.com              
///////////////////////////////////////////////////////////////////////////////

`timescale 1ns/1ps

module logger

  #(parameter EVENT_SIZE = 64,
  parameter DATA_WIDTH = 64,
  parameter CTRL_WIDTH = 8,
  parameter PACKET_SIZE = 10 
  )
	( 
	//From arbitter
	input  [EVENT_SIZE-1:0]			data_arbitter,
	input 							req_arbitter,

	//From sender
	input  [EVENT_SIZE-1:0]			data_sender0,
	input 							req_sender0,
	
	input  [EVENT_SIZE-1:0]			data_sender1,
	input 							req_sender1,

	input  [EVENT_SIZE-1:0]			data_sender2,
	input 							req_sender2,
	
	input  [EVENT_SIZE-1:0]			data_sender3,
	input 							req_sender3,
	
	// To CPU
    output reg [DATA_WIDTH-1:0]           out_data_mac,
	output reg [CTRL_WIDTH-1:0]						out_ctrl_mac,
    output  reg                        			      out_wr_mac,
    input                            			  out_rdy_mac,
	
	//Timings
	input                              reset,
    input                              clk
	);
	
//------------------buffer--------------

reg [7:0] counter; //counts for packets size reach to PACKET_SIZE 
reg [EVENT_SIZE-1:0] in_data;
	
reg bfr_wr_en;
reg bfr_rd_en;
wire bfr_empty;
wire [EVENT_SIZE-1:0]	bfr_out_data;

logger_small_fifo
    #( .WIDTH(EVENT_SIZE),
       .MAX_DEPTH_BITS(2),
       .NEARLY_FULL(3))
 logger_small_fifo
    ( 
     .din				(in_data),
     .wr_en				(bfr_wr_en),
     .rd_en				(bfr_rd_en), 
     
     .dout    			(bfr_out_data),

     .empty				(bfr_empty),
     
     .reset				(reset),
     .clk				(clk)
     );

	 
	 
	 
	 
//-------------------------------------------------------------------	 
reg [EVENT_SIZE-1:0] data_sender_reg0;
reg [EVENT_SIZE-1:0] data_sender_reg1;
reg [EVENT_SIZE-1:0] data_sender_reg2;
reg [EVENT_SIZE-1:0] data_sender_reg3;
reg [EVENT_SIZE-1:0] data_arbitter_reg;


always @(posedge clk)
begin
if (reset) begin
	data_sender_reg0 <= 0;
	data_sender_reg1 <= 0;
	data_sender_reg2 <= 0;
	data_sender_reg3 <= 0;
	data_arbitter_reg <= 0;
end else begin
	if (req_sender0) data_sender_reg0 <= data_sender0;
	if (req_sender1) data_sender_reg1 <= data_sender1;
	if (req_sender2) data_sender_reg2 <= data_sender2;
	if (req_sender3) data_sender_reg3 <= data_sender3;
	if (req_arbitter) data_arbitter_reg <= data_arbitter;
end
end

reg req_sender0_reg;
reg req_sender1_reg;
reg req_sender2_reg;
reg req_sender3_reg;
reg req_arbitter_reg;
always @(posedge clk) begin
if (reset) begin
	req_sender0_reg <= 0;
	req_sender1_reg <= 0;
	req_sender2_reg <= 0;
	req_sender3_reg <= 0;
	req_arbitter_reg <= 0;

end else begin
	req_sender0_reg <= req_sender0;
	req_sender1_reg <= req_sender1;
	req_sender2_reg <= req_sender2;
	req_sender3_reg <= req_sender3;
	req_arbitter_reg <= req_arbitter;
end
end

reg [4:0] req;
reg [4:0] req_s;
reg [4:0] req_p;
always @(posedge clk) begin
	if (reset) begin 
		req_s = 'b00000;
		bfr_wr_en <= 0;
	end else begin

		
		bfr_wr_en <= |req;
		req_s = 'b00000;
		
		if (req[0]) begin
			req_s[0] = 1;
			in_data <= data_arbitter_reg;
		end else if (req[1]) begin
			req_s[1] = 1;
			in_data <= data_sender_reg0;
		end else if (req[2]) begin
			req_s[2] = 1;
			in_data <= data_sender_reg1;
		end else if (req[3]) begin
			req_s[3] = 1;
			in_data <= data_sender_reg2;
		end else if (req[4]) begin
			req_s[4] = 1;
			in_data <= data_sender_reg3;
		end		
	
end
end
always @* begin
if(reset)  req='b0000;
else begin
		req[0] = req_p[0] ^ req_arbitter_reg ^ req_s[0];
		req[1] = req_p[1] ^ req_sender0_reg ^ req_s[1];
		req[2] = req_p[2] ^ req_sender1_reg ^ req_s[2];
		req[3] = req_p[3] ^ req_sender2_reg ^ req_s[3];
		req[4] = req_p[4] ^ req_sender3_reg ^ req_s[4];		
end
end
always @(posedge clk) begin
if (reset) req_p <= 0;
else req_p <= req;
end

	//read from buffer and send them

wire my_wr_en;
assign my_wr_en = !bfr_empty & !out_wr_mac;
always @(posedge clk) begin
if (reset) begin
	out_wr_mac <= 0;
	counter <= 0;
	bfr_rd_en <= 0;
end else begin

	// out_data_mac [DATA_WIDTH-1:EVENT_SIZE]	= 0;
	bfr_rd_en <= my_wr_en && |counter;
	out_wr_mac <= my_wr_en;
		
		out_data_mac [EVENT_SIZE-1:0] <= bfr_out_data;
		if (counter==0)	out_data_mac <= 'h8000006400020190;// nothing importatnt
		
		out_ctrl_mac <= 0;
		if (counter==0)	out_ctrl_mac <= 'hff;
		else begin 
			if (counter==PACKET_SIZE) out_ctrl_mac <= 'h01;
		end
		
		
	counter <= counter+out_wr_mac;
	if ((counter == PACKET_SIZE) & out_wr_mac) counter <= 0;

end
end

endmodule 
